#include "nestypes.h"
#include "logtable.h"
#include "s_sng.h"

#if USE_MAME_ENGINE


#include "mame/sn76496.h"

void SNGSOUND::Reset(void)
{
	if (!pSynth) return;
	SN76496_init((SN76496 *)pSynth, iClock, iFrequency);
}
void SNGSOUND::WriteStereo(Uint8 value)
{
	if (!pSynth) return;
	SN76496WriteStereo((SN76496 *)pSynth, value);
}
void SNGSOUND::WriteData(Uint8 value)
{
	if (!pSynth) return;
	SN76496Write((SN76496 *)pSynth, value);
}
void SNGSOUND::Synth(Int16 *pBuffer, Uint32 numSamples)
{
	if (!pSynth) return;
	SN76496Update_16((SN76496 *)pSynth, pBuffer, numSamples);
}
SNGSOUND::SNGSOUND(int frequency, int clock)
{
	iClock = clock;
	iFrequency = frequency;
	pSynth = new SN76496;
	if (!pSynth) return;
	Reset();
}

SNGSOUND::~SNGSOUND()
{
	if (!pSynth) return;
	delete pSynth;
}


#else

#define CPS_SHIFT 12
#define LOG_KEYOFF (32 << LOG_BITS)

#define FB_WNOISE   0x14002
#define FB_PNOISE   0x08000
#define NG_PRESET   0x0F35

static SNGSOUND sng;

#define V(a) (((a * (1 << LOG_BITS)) / 3) << 1)
static Uint32 voltbl[16] = {
	V(0x0), V(0x1), V(0x2),V(0x3),V(0x4), V(0x5), V(0x6),V(0x7),
	V(0x8), V(0x9), V(0xA),V(0xB),V(0xC), V(0xD), V(0xE),LOG_KEYOFF
};
#undef V

Int32 SNGSOUND::SNGSoundSquare(SNG_SQUARE *ch)
{
	Uint32 output = LOG_KEYOFF;
	if (ch->update)
	{
		if (ch->update & 1)
		{
			ch->spd = ch->regs[0] << CPS_SHIFT;
		}
		if (ch->update & 2)
		{
			ch->vol = voltbl[ch->regs[1] & 0xF];
		}
		ch->update = 0;
	}

	if (ch->spd >= (0x10 << CPS_SHIFT))
	{
		ch->cycles -= ch->cps;
		while (ch->cycles < 0)
		{
			ch->cycles += ch->spd;
			ch->adr++;
		}

		if (!ch->mute && (ch->adr & 1))
		{
			output = ch->vol + mastervolume;
		}
	}
	else
	{
		output = ch->vol + mastervolume;
	}
	return LogToLinear(output, LOG_LIN_BITS - 21);
}


Uint32 SNGSOUND::SNGSoundNoise(SNG_NOISE *ch)
{
	Uint32 output = LOG_KEYOFF;
	if (ch->update)
	{
		if (ch->update & 1)
		{
			ch->spd = ch->regs[0] << CPS_SHIFT;
		}
		if (ch->update & 2)
		{
			ch->vol = voltbl[ch->regs[1] & 0xF];
		}
		ch->update = 0;
	}

	if (ch->spd >= (1 << CPS_SHIFT))
	{
		ch->cycles -= ch->cps;
		while (ch->cycles < 0)
		{
			ch->cycles += ch->spd;
			if (ch->rng & 1) ch->rng ^= ch->fb;
			ch->rng >>= 1;
		}
		if (!ch->mute && (ch->rng & 1))
		{
			output = ch->vol + mastervolume;
		}
	}

	return LogToLinear(output, LOG_LIN_BITS - 21);
}


#define VOL_SFT 8

void SNGSOUND::Synth(Int16 *pBuffer, Uint32 numSamples)
{
	Uint32 i;
	for (i = 0; i < numSamples; i++)
	{
		Int32 accum;
		pBuffer[0] = pBuffer[1] = 0;
		accum = SNGSoundSquare(&square[0]) >> VOL_SFT;
		if (ggs & 0x01) pBuffer[0] += accum;
		if (ggs & 0x10) pBuffer[1] += accum;
		accum = SNGSoundSquare(&square[1]) >> VOL_SFT;
		if (ggs & 0x02) pBuffer[0] += accum;
		if (ggs & 0x20) pBuffer[1] += accum;
		accum = SNGSoundSquare(&square[2]) >> VOL_SFT;
		if (ggs & 0x04) pBuffer[0] += accum;
		if (ggs & 0x40) pBuffer[1] += accum;
		accum = SNGSoundNoise(&noise) >> VOL_SFT;
		if (ggs & 0x08) pBuffer[0] += accum;
		if (ggs & 0x80) pBuffer[1] += accum;
		pBuffer += 2;
	}
}

void SNGSOUND::WriteData(Uint8 value)
{
	Uint ch;
	if (first && value & 0x80) WriteData(0);
	if (first)
	{
		ch = (first & 0x60) >> 5;
		square[ch].regs[0] = ((value & 0x3F) << 4) + (first & 0xF);
		square[ch].update |= 1 << 0;
		if (ch == 2 && noise.mode == 3)
		{
			noise.regs[0] = (square[2].regs[0] + 1) << 1;
			noise.update |= 1 << 0;
		}
		first = 0;
	}
	else
	{
		switch (value & 0xF0)
		{
			case 0x80:	case 0xA0:	case 0xC0:
				first = value;
				break;
			case 0x90:	case 0xB0:	case 0xD0:
				ch = (value & 0x60) >> 5;
				square[ch].regs[1] = value & 0xF;
				square[ch].update |= 1 << 1;
				break;
			case 0xE0:
				noise.mode = value & 0x7;
				if (noise.mode == 3)
				{
					noise.fb = FB_WNOISE;
					noise.regs[0] = (square[2].regs[0] + 1) << 1;
					noise.update |= 1 << 0;
				}
				else
				{
					noise.fb = (noise.mode & 4) ? FB_WNOISE : FB_PNOISE;
					noise.regs[0] = 1 << (1 + (noise.mode & 3));
					noise.update |= 1 << 0;
				}
				noise.rng = NG_PRESET;
				break;
			case 0xF0:
				noise.regs[1] = value & 0xF;
				noise.update |= 1 << 1;
				break;
		}
	}
}

void SNGSOUND::WriteStereo(Uint8 value)
{
	ggs = value;
}

static Uint32 DivFix(Uint32 p1, Uint32 p2, Uint32 fix)
{
	Uint32 ret;
	ret = p1 / p2;
	p1  = p1 % p2;/* p1 = p1 - p2 * ret; */
	while (fix--)
	{
		p1 += p1;
		ret += ret;
		if (p1 >= p2)
		{
			p1 -= p2;
			ret++;
		}
	}
	return ret;
}

void SNGSOUND::Reset(void)
{
	int i;

	for (i = 0; i < 3; i++)
	{
		square[i].mute = 0;
		square[i].update = 0;
		square[i].vol = LOG_KEYOFF;
		square[i].cycles = 0;
		square[i].regs[0] = 0;
		square[i].regs[1] = 0;
		square[i].adr = 0;
	}
	noise.update = 0;
	noise.vol = LOG_KEYOFF;
	noise.cycles = 0;
	noise.regs[0] = 0;
	noise.regs[1] = 0;
	noise.rng = NG_PRESET;
	noise.mode = 0;

	first = 0;

	WriteData(0xFF);
	WriteData(0xE4);

	WriteData(0x9F);
	WriteData(0x80);
	WriteData(0x00);

	WriteData(0xBF);
	WriteData(0xA0);
	WriteData(0x00);

	WriteData(0xDF);
	WriteData(0xC0);
	WriteData(0x00);

	WriteStereo(0xFF);
}

SNGSOUND::SNGSOUND(int frequency, int clock)
{
	int i;
	LogTableInitialize();
	for (i = 0; i < 3; i++)
	{
		XMEMSET(&square[i], 0, sizeof(SNG_SQUARE));
		square[i].cps = DivFix(clock, 16 * frequency, CPS_SHIFT);
	}
	XMEMSET(&noise, 0, sizeof(SNG_NOISE));
	noise.cps = DivFix(clock, 16 * frequency, CPS_SHIFT);
	mastervolume = 0;

	Reset();
}

SNGSOUND::~SNGSOUND()
{
}

#endif
