#include "s_vrc7.h"


static Uint8 optbl[9] = 
{
	0x00, 0x01, 0x02, 0x08, 0x09, 0x0A, 0x10, 0x11, 0x12
};

static Uint8 romtone[1][16 * 19] = 
{
	{
#include "i_fmunit.h"
	}
};


void OPLLSOUND::opll_writeinst(Uint nch, Uint8 *tone)
{
	OPLWriteReg(ym3812, 0x20 + optbl[nch], tone[0]);
	OPLWriteReg(ym3812, 0x23 + optbl[nch], tone[1]);
	OPLWriteReg(ym3812, 0x60 + optbl[nch], tone[4]);
	OPLWriteReg(ym3812, 0x63 + optbl[nch], tone[5]);
	OPLWriteReg(ym3812, 0x80 + optbl[nch], tone[6]);
	OPLWriteReg(ym3812, 0x83 + optbl[nch], tone[7]);
	OPLWriteReg(ym3812, 0xe0 + optbl[nch], (tone[3] >> 3) & 0x01);
	OPLWriteReg(ym3812, 0xe3 + optbl[nch], (tone[3] >> 4) & 0x01);
	OPLWriteReg(ym3812, 0xc0 + nch, (tone[3] << 1) & 0x0e);
	ch[nch].mtl = (tone[8] ^ 0x80) + (tone[2] & 0x3f);
	ch[nch].ctl = (tone[9] ^ 0x80);
	ch[nch].mks = tone[2] & 0xC0;
	ch[nch].cks = tone[3] & 0xC0;
}

void OPLLSOUND::opll_writevol(Uint nch)
{
	Uint value, mtl, ctl;
	value = ch[nch].volinst;
	if ((rhythmc & 0x20) && nch == 6)
	{
		mtl = ch[nch].mtl;
		ctl = ((value & 0x0f) << 2) + ch[nch].ctl + mastervolume;
		mtl = (mtl < 0x80) ? 0 : ((mtl > 0xbf) ? 0x3f : (mtl & 0x3f));
		ctl = (ctl < 0x80) ? 0 : ((ctl > 0xbf) ? 0x3f : (ctl & 0x3f));
		OPLWriteReg(ym3812, 0x40 + optbl[nch], ch[nch].mks | mtl);
		OPLWriteReg(ym3812, 0x43 + optbl[nch], ch[nch].cks | ctl);
	}
	else if ((rhythmc & 0x20) && nch > 6)
	{
		mtl = ((value & 0xf0) >> 2) + ch[nch].mtl + mastervolume;;
		ctl = ((value & 0x0f) << 2) + ch[nch].ctl + mastervolume;;
		mtl = (mtl < 0x80) ? 0 : ((mtl > 0xbf) ? 0x3f : (mtl & 0x3f));
		ctl = (ctl < 0x80) ? 0 : ((ctl > 0xbf) ? 0x3f : (ctl & 0x3f));
		OPLWriteReg(ym3812, 0x40 + optbl[nch], ch[nch].mks | mtl);
		OPLWriteReg(ym3812, 0x43 + optbl[nch], ch[nch].cks | ctl);
	}
	else
	{
		opll_writeinst(nch, inst[value >> 4]);
		mtl = ch[nch].mtl;
		ctl = ((value & 0x0f) << 2) + ch[nch].ctl + mastervolume;;
		mtl = (mtl < 0x80) ? 0 : ((mtl > 0xbf) ? 0x3f : (mtl & 0x3f));
		ctl = (ctl < 0x80) ? 0 : ((ctl > 0xbf) ? 0x3f : (ctl & 0x3f));
		OPLWriteReg(ym3812, 0x40 + optbl[nch], ch[nch].mks | mtl);
		OPLWriteReg(ym3812, 0x43 + optbl[nch], ch[nch].cks | ctl);
	}
}

void OPLLSOUND::opll_writefreq(Uint nch)
{
	Uint value;
	value = (ch[nch].freqh << 8) + ch[nch].freql;
	if ((rhythmc & 0x20) && nch >= 6)
	{
		OPLWriteReg(ym3812, 0xA0 + nch, (value << 1) & 0xFE);
		OPLWriteReg(ym3812, 0xB0 + nch, (value >> 7) & 0x5F);
	}
	else
	{
		OPLWriteReg(ym3812, 0xA0 + nch, (value << 1) & 0xFE);
		OPLWriteReg(ym3812, 0xB0 + nch, (value >> 7) & 0x7F);
	}
}

void OPLLSOUND::opll_write(Uint address, Uint value)
{
	switch (adr & 0x38)
	{
		case 0x00:
			inst[0][adr & 0x07] = value;
			break;
		case 0x08:
			if ((adr & 0x0F) == 0x0E)
			{
				if ((value & 0x20) && !(rhythmc & 0x20))
				{
					rhythmc = value;
					opll_writeinst(6, inst[16]);
					opll_writevol(6);
					opll_writefreq(6);
					opll_writeinst(7, inst[17]);
					opll_writevol(7);
					opll_writefreq(7);
					opll_writeinst(8, inst[18]);
					opll_writevol(8);
					opll_writefreq(8);
				}
				else
				{
					rhythmc = value;
				}
				OPLWriteReg(ym3812, 0xBD, rhythmc & 0x3F);
			}
			break;
		case 0x18:
			if ((adr & 0x0F) >= 9) break;
		case 0x10:
			ch[adr & 0x0F].freql = value;
			opll_writefreq(adr & 0x0F);
			break;
		case 0x28:
			if ((adr & 0x0F) >= 9) break;
		case 0x20:
			ch[adr & 0x0F].freqh = value;
			opll_writefreq(adr & 0x0F);
			break;
		case 0x38:
			if ((adr & 0x0F) >= 9) break;
		case 0x30:
			ch[adr & 0x0F].volinst = value;
			opll_writevol(adr & 0x0F);
			break;
	}
}

__inline Uint32 GetDwordLE(Uint8 *p)
{
	return p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
}

void OPLLSOUND::opll_setinst(Uint8 *p)
{
	Uint i, j;
	if ((GetDwordLE(p) & 0xf0ffffff) == 0x304c4c49)
	{
		OPLSetExtraParameter(p + 4);
		p += 8;
		for(j = 8; j < 16; j++) inst[0][j] = *p++;
		/* new version */
		for(i = 1; i < 19; i++)
		{
			for(j = 0; j < 16; j++) inst[i][j] = *p++;
		}
	}
	else
	{
		Uint8 initdata[4] = {0,0,0,0};
		OPLSetExtraParameter(initdata);
		for(j = 8; j < 16; j++) inst[0][j] = 0;
		for(i = 1; i < 16; i++)
		{
			for(j = 0; j < 8; j++) inst[i][j] = *p++;
			for(j = 8; j < 16; j++) inst[i][j] = 0;
		}
	}
}

void OPLLSOUND::opll_reset(void)
{
	Uint i;
	OPLResetChip(ym3812);
    for(i = 0; i < 9; i++) XMEMSET(&ch[i], 0, sizeof(OPLL_CH));
    for(i = 0; i < 0x100; i++) OPLWriteReg(ym3812, i, 0x00);
	adr = 0;
	rhythmc = 0x00;
	mastervolume = 0;
	OPLWriteReg(ym3812, 0xBD, rhythmc);
	OPLWriteReg(ym3812, 0x01, 0x20);
    for(i = 0; i < 0x40; i++) opll_write(i, 0x00);
}

#if 0
Int32 OPLLSOUND::SynthSample(void)
{
	INT16 buf;
	if (!ym3812) return 0;
	YM3812UpdateOne(ym3812, &buf, 1);
	return buf << 8;
}
#endif

void OPLLSOUND::Synth(Int16 *pBuffer, Uint32 numSamples)
{
	Uint32 i;
	if (!ym3812) return;
	YM3812UpdateOne(ym3812, pBuffer + numSamples, numSamples);
	/* stereo */
	for (i = 0; i < numSamples; i++)
	{
		Int16 s;
		s = pBuffer[numSamples + i];
		pBuffer[i * 2 + 0] = s;
		pBuffer[i * 2 + 1] = s;
	}
}

void OPLLSOUND::WriteAddress(Uint8 value)
{
	adr = value;
}

void OPLLSOUND::WriteData(Uint8 value)
{
	if (!ym3812) return;
	opll_write(adr, value);
}


void OPLLSOUND::Reset(void)
{
	if (!ym3812) return;
	opll_reset();
}

void VRC7SetTone(Uint8 *p)
{
	if ((GetDwordLE(p) & 0xf0ffffff) == 0x304c4c49)
		XMEMCPY(romtone[0], p, 16 * 19);
	else
		XMEMCPY(romtone[0], p, 8 * 15);
}


OPLLSOUND::OPLLSOUND(int frequency, int clock)
{
	opll_setinst(romtone[0]);
	ym3812 = OPLCreate(OPL_TYPE_YM3812, clock, frequency);
	if(!ym3812) return;
	opll_reset();
}

OPLLSOUND::~OPLLSOUND()
{
	if (ym3812)
	{
		OPLDestroy(ym3812);
		ym3812 = 0;
	}
}
